import os
import logging


# ------------------------------------------------------------------------------
# -- Define the logger for fracture
log = logging.getLogger('fracture')


# ------------------------------------------------------------------------------
# -- This is the location we look to for all the built in plugins, that is
# -- to say plugins which are 'always on'
BUILT_IN_PLUGIN_DIR = os.path.join(
    os.path.dirname(__file__),
    'built-in',
)



"""
# Overview

Fracture is a lightweight and flexible data management system. It allows
you to interact with data through a trait compositing mechanism, whilst
also exposing the ability to quickly query and access information
about the data you're exploring.

# How it works

You start by creating a fracture.Project. The project file is where all
the metadata and look up tables are stored - allowing you to easily search
for data assets as well as find changes.

Fracture comes built-in with a file searching mechanism, but you can extend
this with your own search mechanisms too. For instance, if you have data
on an FTP, or within Source Control and you want to add that data to the
project without having to have it physically on disk you're able to do so
by implementing a fracture.ScanProcess plugin.

Finally, and probably the most important is the DataElement. This is a class
which you can use to express the functionality of data. Rather than having a
1:1 relationship between a DataElement class and a data type the DataElement
class supports class compositing. This allows for a piece of data to be
represented by more than one class simultaneously.

# Examples

This example uses the ```dino_example``` data which you can download from
https://github.com/mikemalinowski/fracture. 

To start with, we create a fracture project. To do this we must specify
two pieces of information, the first being where we want to save our 
project file and the second being the locations where we want fracture
to look for Scan and Data Plugins.

```python
import os
import fracture

project = fracture.create(
    project_path=os.path.join(current_path, '_dino.fracture'),
    plugin_locations=[os.path.join(current_path, 'plugins')]
)
```

This returns a fracture.Project instance which we can then start interacting
with, for instance we can define locations where the project should start
looking for data:

```python
# -- Tell the project where to look for data
project.add_scan_location('/usr/my_data'))
```

Finally, with the project made, and at least one search location added we
can initiate a search...

```python
# -- Now we initiate a scan. This will cycle over all the
# -- scan locations and scrape them for data
project.scan()
```

Scanning is the process of running over all the scan plugins - of which there
is always at least one (the file scraper), and populating the project with
information about each piece of data which is found. The process is pretty
quick and the amount of data stored is minimal - primarily just the identifier
such as the path along with any tags as defined by any DataElement composites
which can represent that data.

With the project populated we can now start querying the project for
data

```python
# -- Now we have scanned we can start to run queries over data
# -- very quickly. We can search by tags, or use the special
# -- * wildcard
for item in project.find('*'):

    # -- By default we get string identifiers back from a find, as
    # -- this is incredibly quick. However, we can ask for the data
    # -- composite representation of the item. A data composite is
    # -- a composition of all the data plugins which can represent
    # -- this identifier.
    item = project.get(item)

    # -- Print the identifier, and the item (which also shows the
    # -- class composition)
    print(item.identifier())
    print('\t%s' % item)

    # -- We can start interacting with this data, calling
    # -- functionality will return a dictionary of all the
    # -- functionality exposed by all the data plugins representing
    # -- this item
    for k, v in item.functionality().items():
        print('\t\t%s = %s' % (k, v))
```

The process of querying is very quick, even for reasonably large data sets. In
the example above we're then asking the project to 'get' the item. This process
take the identifier and binds all the relevent DataElements together which
can possibly represent the data.

Binding is particularly useful when there is no obvious hierarchy between
two elements. For instance, in the ```dino_example``` data set we have a
trait which is ```carnivore``` and a trait which is ```herbivore```. There
is no hierarchical relationship between the two, but an omnivore would need
both. By using class compositing we avoid complex multi-inheritence situations.

Using this same mechanism, if we know the locator of a piece of information,
such as a file path, we can get the composited class directly without having
to run a query, as shown here:

```python
# -- We do not have to utilise the find method to get access to data,
# -- and in fact we can get a Composite representation of data even
# -- if that data is not within our scan location.
data = project.get('/usr/my_data/my_file.txt')
```

For a full demo, download the ```dino_example``` and run main.py

# Data Composition

As mentioned in the examples, we use class composition to bind traits together
to represent data. This means we can have small, self contained traits which
do not need rigid hierarchical structures designed for them.

There are three main composited methods in the DataElement class, specifically:

* label : The first call that returns a positive result is taken
* mandatory_tags : All the lists are collated from all compositions and made unique
* functionality : All dictionaries are combined into a single dictionary
* icon : The first call that returns a positive result is taken

Given the ```dino_example``` files, the velociraptor.png file, when passed
to ```project.get('/usr/my_data/.../velociraptor.png')``` is expressed
as a class formed of the following traits: [Carnivore; File; Image;] where each
trait can expose its own information.

An implementation of a DataElement plugin looks like this:

```python
import re
import fracture

# -- All plugins must inherit from the fracture.DataElement class in order
# -- to be dynamically picked up.
class CarnivoreTrait(fracture.DataElement):

    # -- The data type is mandatory, and is your way of
    # -- denoting the name of this plugin
    data_type = 'carnivore'

    # -- These two lines are not at all required and are here
    # -- just to make performance better
    _split = re.compile('/|\.|,|-|:|_', re.I)
    _has_trait = re.compile('(carnivore|omnivore).*\.', re.I)

    # --------------------------------------------------------------------------
    # -- This method must be re-implemented, and its your oppotunity to
    # -- decide whether this plugin can viably represent the given data
    # -- identifier.
    # -- In this example we use a regex check, but it could be anything
    # -- you want. The key thing to remember is that this is called a lot,
    # -- so performance is worth keeping in mind.
    @classmethod
    def can_represent(cls, identifier):
        if CarnivoreTrait._has_trait.search(identifier):
            return True
        return False

    # --------------------------------------------------------------------------
    # -- This is your way of exposing functionality in a common and consistent
    # -- way. If you know the data types you can of course call things directly
    # -- but this is a good catch up consistent way of exposing functionality
    # -- and is typically harnessed by user interfaces.
    def functionality(self):
        return dict(
            feed_meat=self.feed_meat,
            ),
        )

    # --------------------------------------------------------------------------
    # -- This should return a 'nice name' for the identifier
    def label(self):
        return os.path.basename(self.identifier())

    # --------------------------------------------------------------------------
    # -- As fracture heavily utilises tags, this is your way of defining a
    # -- set of tags which are mandatory for anything with this trait
    def mandatory_tags(self):
        return ['carnivore', 'meat', 'hunter']

    # --------------------------------------------------------------------------
    # -- This is here just as a demonstration of a callable function which
    # -- which can be accessed on the trait
    def feed_meat(self):
        print('Would feed this creature some meat...')
```

By placing a trait plugin anywhere within the plugin locations you define
for your project will immediately make it accessible.


## ScanProcess

By default fracture comes with one built-in scan plugin which handles file
scanning, so that is a good example when wanting to write your own - if you 
have need to do so.

This plugin type defines how to find data. If your data is files on a disk
such as those in the example above then your scan plugin may do little more
than cycle directories and yield file paths.

Alternatively if you're caching data from a REST API you might be utilising
requests within the scan process and feeding back URL's.

# Origin

This library is a variation on the tools demonstrated during
GDC2018 (A Practical Approach to Developing Forward-Facing Rigs, Tools and
Pipelines), which can be explored in more detail here:
https://www.gdcvault.com/play/1025427/A-Practical-Approach-to-Developing

Slide 55 onward explores this concept. It is also explored in detail
on this webpage:
https://www.twisted.space/blog/insight-localised-asset-management

"""
from ._project import icon
from ._project import Project
from ._scan import ScanProcess
from ._element import DataElement

from ._access import create
from ._access import load
from .constants import log



"""
This module holds the abstract class which defines the structure for a
scan plugin.
"""


# ------------------------------------------------------------------------------
# noinspection PyUnusedLocal
class ScanProcess(object):
    """
    This type defines how data is found. This may be done by cycling folders
    and files or it could be scraping a REST API or pulling information from
    a third party source. 
    """
    scan_type = ''

    IS_VALID = 0
    NOT_VALID = 1
    UNKNOWN = 2

    # --------------------------------------------------------------------------
    @classmethod
    def can_represent(cls, location):
        """
        This is used to allow the plugin to determine if it thinks it is
        able to scrape the given data source identifier or not.

        :param location: Location Identifier. This could be a url, or a filepath
            or a uuid etc.
        :type location: string

        :return: bool
        """
        return False

    # --------------------------------------------------------------------------
    @classmethod
    def identifiers(cls, location, skip_pattern, recursive=True):
        """
        This will return a list of data identifiers found in the given data 
        locations.

        NOTE: This should always yield results!

        :param location: Location to scan
        :type location: str

        :param skip_pattern: This is a regular expression object which, if
            matched on a location should be skipped.
        :type skip_pattern: regex

        :param recursive: If True all locations below the given location
            will also be scanned, otherwise only the immediate location will
            be scanned
        :type recursive: True

        :return: generator
        """
        return list()

    # --------------------------------------------------------------------------
    @classmethod
    def above(cls, location):
        """
        Returns all the search locations above the current location. If there is
        no location above, or the concept of 'above' makes no sense for your
        scan plugin you may return an empty list.

        :return:
        """
        return list()

    # --------------------------------------------------------------------------
    @classmethod
    def below(cls, location):
        """
        This will return all the search locations immediately below the given
        location. If there is no location above, or the concept of 'above'
        makes no sense for your scan plugin you may return an empty list.

        :param location:
        :return:
        """
        return list()

    # --------------------------------------------------------------------------
    @classmethod
    def check(cls, identifier):
        """
        Given an identifier, this should check whether the identifier is
        still considered valid or not - or whether it cannot make that
        assumption. 

        This should return one of the three following:

            ScanProcess.IS_VALID
                If the identifier is still active and valid, this should
                be returned

            ScanProcess.NOT_VALID
                If the identifier should no longer be considered within
                fracture this should be returned

            ScanProcess.UNKNOWN
                If the identifier is not digestable by this process for
                any reason this should be returned.

        :return: int 
        """
        return cls.UNKNOWN



"""
These are the primary entry points which we expect users to
access and interact with fracture projects.
"""
from . import _project


# ------------------------------------------------------------------------------
def create(project_path, plugin_locations):
    """
    This will generate a new fracture project and return the project
    instance.

    :param project_path: Path to the location you want to save the project
        path.
    :type project_path: str

    :param plugin_locations: List of locations to search for plugins

    :return: fracture.Project instance
    """
    if not project_path.endswith('fracture'):
        project_path += '.fracture'

    project = load(project_path)
    project.create()

    for location in plugin_locations:
        project.add_plugin_location(location)

    return project


# ------------------------------------------------------------------------------
def load(project_path):
    """
    Loads the given project and returns a Project Instance
    
    :param project_path: Path to the location you want to save the project
        path.
    :type project_path: str

    :return: fracture.Project instance
    """
    return _project.Project(project_path)



"""
This module contains the DataElement class. This is the class which
all data elements inherit from and their compositions are utilised to
represent data.
"""
import xcomposite


# ------------------------------------------------------------------------------
# noinspection PyUnusedLocal
class DataElement(xcomposite.Composition):
    """
    A DataElement is a block of functionality and an interface to a
    piece of data. A piece of data can be represented by multiple
    DataElements in a composite form.
    """

    # -- This allows you to define a unique identifier for the type
    # -- of plugin this is. When querying a return from fracture this will
    # -- form part of the composition string.
    data_type = ''

    # -- This is useful if you plan to re-implement a plugin with the same
    # -- data_type as another plugin. In this way only the highest version
    # -- will be used.
    version = 1

    # -- Priority to determines the order of class composition. A higher
    # -- values means that the class will be composited first. This can
    # -- have an effect on methods decorated with xcomposite.First etc.
    priority = 1

    # --------------------------------------------------------------------------
    def __init__(self, identifier, db):
        super(DataElement, self).__init__()

        self.db = db
        self._identifier = identifier

    # --------------------------------------------------------------------------
    def __repr__(self):
        base_repr = super(DataElement, self).__repr__()
        return base_repr.replace(
            'DataElement',
            'DataElement: %s' % (
                self._identifier,
            )
        )

    # --------------------------------------------------------------------------
    # -- The following 8 methods below should be re-implemented by your
    # -- plugin. You can however omit the decoration of these functions

    # --------------------------------------------------------------------------
    @classmethod
    def can_represent(cls, identifier):
        """
        Determines whether this data plugin can represent the given identifier

        :param identifier: Data Identifier. This could be a url, or a filepath
            or a uuid etc.
        :type identifier: string

        :return: bool
        """
        return False

    # --------------------------------------------------------------------------
    @xcomposite.first_true
    def icon(self):
        """
        This will return the icon for the DataElement. If multiple DataElements
        are bound to represent a piece of information then the first to return
        a non-false result will be taken.

        :return: string 
        """
        return None

    # --------------------------------------------------------------------------
    @xcomposite.first_true
    def label(self):
        """
        Returns the label or name for the DataElement. This is typically 
        considered a 'pretty' name, or short name and does not need to be
        unique.

        If multiple DataElements are bound to represent a piece of information
        then the first to return a non-false result will be taken.

        :return: string 
        """
        return None

    # --------------------------------------------------------------------------
    @xcomposite.extend_unique
    def mandatory_tags(self):
        """
        Returns any tags that should always be assigned to this data element.
        In a situation where multiple DataElements are bound then the combined
        results of all the mandatory tags are used.

        :return: list(str, str, ...)
        """
        return None

    # --------------------------------------------------------------------------
    @xcomposite.update_dictionary
    def functionality(self):
        """
        This exposes per-data functionality in the form of a dicitionary where
        the key is the string accessor, and the value is the callable.

        In a situation where multiple DataElements are bound then a single
        dictionary with all the entries combined is returned.

        :return: dict(label: callable,)
        """
        return xcomposite.Ignore

    # --------------------------------------------------------------------------
    # -- NO METHODS BELOW NEED TO BE RE-IMPLEMENTED

    # --------------------------------------------------------------------------
    def identifier(self):
        """
        This returns the identifier of this data element.

        :return: string 
        """
        return self._identifier

    # --------------------------------------------------------------------------
    def tags(self):
        """
        Returns a list of all the tags assigned to this DataElement

        :return: None 
        """
        return self.db.tags(self._identifier)

    # --------------------------------------------------------------------------
    def tag(self, tags):
        """
        Assigns the given tags to this data element

        :param tags: This can be a single string tag, or a list of tags

        :return: None 
        """
        return self.db.tag(self._identifier, tags)

    # --------------------------------------------------------------------------
    def untag(self, tags):
        """
        Un-assigns the given tags from this data element

        :param tags: This can be a single string tag, or a list of tags

        :return: None 
        """
        return self.db.untag(self._identifier, tags)



import os
import setuptools

short_description = 'Fracture is a lightweight and flexible data management system'
if os.path.exists('README.md'):
    with open('README.md', 'r') as fh:
        long_description = fh.read()

else:
    long_description = short_description

setuptools.setup(
    name='fracture',
    version='0.9.1',
    author='Mike Malinowski',
    author_email='mike@twisted.space',
    description=short_description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mikemalinowski/fracture',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=['factories'],
    package_data={
        '': ['sql/*.sql', '_res/*.png'],
    },
    keywords="fracture data composite",
)



"""
The fracture.Project class is the central core to Fracture. It handles all the
database io and factories.
"""
import os
import re
import sys
import json
import sqlite3
import factories

from ._utils import Signal
from ._scan import ScanProcess
from ._element import DataElement

from . import constants


# ------------------------------------------------------------------------------
class Project(object):
    """
    The Project is the main backbone of fracture. This handles all the
    data accessing and getting. It also manages all the factories and
    general functionality.
    """

    # -- This is the location where we store all the
    # -- sqlite files
    _COMMAND_DIR = os.path.join(
        os.path.dirname(__file__),
        'sql',
    )

    _DEBUG = False

    # --------------------------------------------------------------------------
    def __init__(self, identifier):
        self.identifier = identifier

        # -- This signal is triggered on each scan identifier
        self.scanned = Signal()

        # -- This signal is emitted whenever a scan is completed
        self.scan_complete = Signal()

        # -- Read in all our command data. We do this once so we dont take
        # -- an IO hit each time we want to run an sql query
        self._commands = self._command_dict()

        # -- Only request plugin locations from the db if the
        # -- db actually exists
        plugin_locations = list()
        if os.path.exists(identifier):
            plugin_locations = self.plugin_locations()

        # -- Always add our built in plugins
        plugin_locations.append(constants.BUILT_IN_PLUGIN_DIR)

        # -- We do not hard code the concept of scan mechanisms or
        # -- data elements. Instead we utilise dynamic factories which
        # -- load plugins on the fly.
        self._scan_factory = factories.Factory(
            abstract=ScanProcess,
            paths=plugin_locations,
            plugin_identifier='scan_type',
        )

        self._data_factory = factories.Factory(
            abstract=DataElement,
            paths=plugin_locations,
            plugin_identifier='data_type',
        )

        # -- We will access data plugins a lot - and they need to be
        # -- in priority order, therefore we cache the plugin classes
        # -- in a prioritised list. We then re-cache this data whenever
        # -- plugin paths are changed.
        self._data_plugins = sorted(
            self._data_factory.plugins(),
            key=lambda x: x.priority,
            reverse=True,
        )

    # --------------------------------------------------------------------------
    def create(self):
        """
        This should be used to create the datasource (providing you decide
        to support such a feature).
        
        :return: None 
        """
        with ConnectionContext(self.identifier, commit=True) as ctxt:
            self._execute(
                ctxt,
                'create',
            )
            return True

    # --------------------------------------------------------------------------
    def find(self, tags, limit=None):
        """
        This should return a list of identifiers which match the given
        tags. 
        
        :param tags: List of tags to match against. 
        :type tags: list(str, str, )
        
        :param limit: Maximum number of hits to return
        :type limit: int
        
        :return: list(str, str, ...)
        """
        with ConnectionContext(self.identifier, get=True) as ctxt:
            # -- Define our replacements
            replacements = {
                '$(LIMIT)': limit or 9 ** 9,
            }

            if tags:
                if not isinstance(tags, (list, tuple)):
                    tags = [tags]

                # -- We need to build up strings for searching by
                # -- tags and locators
                compare_str = ''
                like_str = ''
                name_str = ''

                for tag in tags:
                    compare_str += "tag='%s' OR " % tag
                    like_str += "identifier LIKE '%" + tag + "%' AND "
                    name_str += "name LIKE '%" + tag + "%' AND "

                compare_str = compare_str[:-4]
                like_str = like_str[:-5]
                name_str = name_str[:-5]

                replacements.update(
                    {
                        '$(TAG_COMPARE)': compare_str,
                        '$(LIKE_COMPARE)': like_str,
                        '$(NAME_COMPARE)': name_str,
                        '$(COMPARE_COUNT)': str(len(tags)),
                    },
                )

                # -- Run the sql statement
                self._execute(
                    ctxt,
                    'find',
                    replacements=replacements,
                )

            else:
                self._execute(
                    ctxt,
                    'find_all',
                    replacements=replacements,
                )

        return [
            row[1]
            for row in ctxt.results
        ]

    # --------------------------------------------------------------------------
    def tag(self, identifier, tags):
        """
        This will assign the given tags to the element with the given 
        identifier.
        
        :param identifier: Data Element identifier
        :type identifier: str
        
        :param tags: A single tag, or list of tags to assign
        :type tags: string or list(str, str, ...)
        
        :return: None 
        """
        with ConnectionContext(self.identifier, commit=True) as ctxt:
            if not isinstance(tags, (list, tuple)):
                tags = [tags]

            for tag in tags:
                self._execute(
                    ctxt,
                    'tags_add',
                    replacements={
                        '$(IDENTIFIER)': identifier,
                        '$(TAG)': tag.lower(),
                    },
                )

    # --------------------------------------------------------------------------
    def untag(self, identifier, tags):
        """
        This will unassign the given tags from the element with the given 
        identifier.
        
        :param identifier: Data Element identifier
        :type identifier: str
        
        :param tags: A single tag, or list of tags to unassign
        :type tags: string or list(str, str, ...)
        
        :return: None 
        """
        with ConnectionContext(self.identifier, commit=True) as ctxt:
            if not isinstance(tags, (list, tuple)):
                tags = [tags]

            for tag in tags:
                self._execute(
                    ctxt,
                    'tags_rem',
                    replacements={
                        '$(IDENTIFIER)': identifier,
                        '$(TAG)': tag,
                    },
                )

    # --------------------------------------------------------------------------
    def tags(self, identifier):
        """
        This returns a list of all the tags which are assigned to the 
        given identifier.
        
        :param identifier: Data Element identifier
        :type identifier: str
        
        :return: list(str, str, ...)
        """
        with ConnectionContext(self.identifier, get=True) as ctxt:
            self._execute(
                ctxt,
                'tags_get',
                replacements={
                    '$(IDENTIFIER)': identifier,
                },
            )

        return [
            str(result[0])
            for result in ctxt.results
        ]

    # --------------------------------------------------------------------------
    def add(self, identifier):
        """
        Given an elements identifier, name, hash and data this will add or 
        update that elements information into the repository.
        
        :param identifier: Data Element identifier
        :type identifier: str
        
        :return: None 
        """
        with ConnectionContext(self.identifier, commit=True) as ctxt:
            self._execute(
                ctxt,
                'add',
                replacements={
                    '$(IDENTIFIER)': identifier,
                },
            )

    # --------------------------------------------------------------------------
    def remove(self, identifier=None):
        """
        This will remove the given identifier from the repository if it 
        exists.
        
        :param identifier: Data Element identifier
        :type identifier: str
        
        :return: None
        """
        with ConnectionContext(self.identifier, commit=True) as ctxt:
            self._execute(
                ctxt,
                'remove',
                replacements={
                    '$(IDENTIFIER)': identifier,
                },
            )

    # --------------------------------------------------------------------------
    def settings(self):
        """
        This will attempt to load the settings. In a situation where no
        settings are found an empty dictionary is returned.

        :return: dict
        """
        with ConnectionContext(self.identifier, get=True) as ctxt:
            self._execute(ctxt, 'settings_get')

        try:
            return json.loads(ctxt.results[0][0])
        except IndexError:
            return dict()

    # --------------------------------------------------------------------------
    def save_settings(self, settings):
        """
        Given a settings dictionary, this will be stored within the fracture
        database. The data being given must be json serialisable.

        :param settings: dict

        :return: None
        """

        with ConnectionContext(self.identifier, commit=True) as ctxt:
            self._execute(
                ctxt,
                'settings_set',
                replacements={'$(SETTINGS)': json.dumps(settings)},
            )
    
    # --------------------------------------------------------------------------
    def scan_locations(self):
        """
        This returns a list of scan locations stored within the repository.
        
        :return: list(str, str, ...) 
        """
        return self.settings().get('scan_locations', list())

    # --------------------------------------------------------------------------
    def add_scan_location(self, location):
        """
        This will store the given location in the repository. When a scan
        is initiated this location will then be picked up.
        
        :param location: Location to add to the repository
        :type location: str
        
        :return: 
        """
        # -- Get the current settings
        settings = self.settings()

        # -- Update the locations, ensuring the data always conforms
        key = 'scan_locations'
        settings[key] = settings.get(key, list()) + [self._conform(location)]

        # -- Save the settings back into the db
        self.save_settings(settings)

    # --------------------------------------------------------------------------
    def remove_scan_location(self, location):
        """
        This will remove the given location from the repositories list of 
        scan locations.
        
        :param location: Location to remove from the repository
        :type location: str
        
        :return: 
        """
        # -- Get the current settings
        settings = self.settings()

        # -- Update the locations, ensuring the data always conforms
        key = 'scan_locations'
        settings[key] = settings.get(key, list()).remove(location)

        # -- Save the settings back into the db
        self.save_settings(settings)

    # --------------------------------------------------------------------------
    def skip_regexes(self):
        """
        This returns a list of regular expression strings stored within 
        the repository.
        
        :return: list(str, str, ...) 
        """
        return self.settings().get('skip_regex', list())

    # --------------------------------------------------------------------------
    def add_skip_regex(self, pattern):
        """
        This will store the given regex pattern in the repository. When a scan
        is initiated this regex will then be picked up and omit and locations
        containing this pattern.
        
        :param pattern: String regex
        :type pattern: str
        
        :return: 
        """
        # -- Get the current settings
        settings = self.settings()

        # -- Update the locations, ensuring the data always conforms
        key = 'skip_regex'
        settings[key] = settings.get(key, list()) + [self._conform(pattern)]

        # -- Save the settings back into the db
        self.save_settings(settings)

    # --------------------------------------------------------------------------
    def remove_skip_regex(self, pattern):
        """
        This will remove the given regex string pattern from the repository.
        
        :param pattern: String regex
        :type pattern: str
        
        :return: 
        """
        # -- Get the current settings
        settings = self.settings()

        # -- Update the locations, ensuring the data always conforms
        key = 'skip_regex'
        settings[key] = settings.get(key, list()).remove(pattern)

        # -- Save the settings back into the db
        self.save_settings(settings)

    # --------------------------------------------------------------------------
    def plugin_locations(self):
        """
        This will return all the places this project is currently
        loading plugins from.

        :return: list(str, str, ...)
        """
        return self.settings().get('plugin_locations', list())

    # --------------------------------------------------------------------------
    def add_plugin_location(self, location):
        """
        Adds the given location to the list of locations being searched
        for when looking for plugins. This data is persistent between
        sessions and will invoke a reload of plugins.

        :param location: Location to add

        :return: None
        """
        # -- Get the current settings
        settings = self.settings()

        # -- Update the locations, ensuring the data always conforms
        key = 'plugin_locations'
        settings[key] = settings.get(key, list()) + [self._conform(location)]

        # -- Save the settings back into the db
        self.save_settings(settings)

        # -- Now update teh factories
        self._scan_factory.add_path(location)
        self._data_factory.add_path(location)

        # -- Re-cache the sorted data plugins
        self._data_plugins = sorted(
            self._data_factory.plugins(),
            key=lambda x: x.priority,
            reverse=True,
        )

    # --------------------------------------------------------------------------
    def remove_plugin_location(self, location):
        """
        This will remove a plugin from the plugin location list and
        trigger a rescan for plugins.

        :param location: Location to remove
        :return:
        """
        # -- Get the current settings
        settings = self.settings()

        # -- Update the locations, ensuring the data always conforms
        key = 'plugin_locations'
        settings[key] = settings.get(key, list()).remove(location)

        # -- Save the settings back into the db
        self.save_settings(settings)

        # -- Update the factories
        self._scan_factory.remove_path(location)
        self._data_factory.remove_path(location)

        # -- Re-cache the sorted data plugins
        self._data_plugins = sorted(
            self._data_factory.plugins(),
            key=lambda x: x.priority,
            reverse=True,
        )

    # --------------------------------------------------------------------------
    def scan(self, locations=None, recursive=True, full=True):
        """
        This will cycle over all the search locations stored in the
        fracture project and attempt to populate the db with that data
        if the data has been changed or is new.

        :return: None 
        """
        # -- Log the start
        constants.log.info('Starting Scan : %s' % locations)

        # -- Get a list of locations to skip, if we have any then
        # -- we compile them into a single regex
        skip_regex = None
        patterns = self.skip_regexes()

        if patterns:
            skip_regex = re.compile(
                '(' + ')|('.join(patterns) + ')',
            )

        # -- Take the given scan locations if the user passes them
        # -- otherwise we take all the scan locations stored within
        # -- the project
        locations = locations or self.scan_locations()
        if not isinstance(locations, (list, tuple)):
            locations = [locations]

        # -- We return a list of all the added identifiers, so
        # -- keep track of that here
        scanned_identifiers = list()

        with ConnectionContext(self.identifier, commit=True) as ctxt:
            # -- Get a list of locations to scan
            for location in locations:

                # -- Cycle over all our scan plugins
                for scan_plugin in self._scan_factory.plugins():

                    # -- Skip any plugins incapable of scanning
                    # -- this location
                    if not scan_plugin.can_represent(location):
                        continue

                    # -- We now need to cycle over all the identifiers given
                    # -- to us via the scan plugin. Note that the scan plugin
                    # -- always yields
                    for identifier in scan_plugin.identifiers(location, skip_regex, recursive=recursive):

                        self._execute(
                            ctxt,
                            'add',
                            replacements={
                                '$(IDENTIFIER)': identifier,
                            },
                        )

                        # -- Emit the scan signal
                        self.scanned.emit(identifier)

                        # -- Store the fact that we have matched this item
                        scanned_identifiers.append(identifier)

        # -- A full scan involves data extraction such as tags,
        # -- which we only want to do if we're performing a full scan
        if full:

            # -- As a performance optimisation we first add all the
            # -- dates to the db, so get a list of all the tags
            all_tags = list()
            mapped_tags = dict()

            for scanned_identifier in scanned_identifiers:
                element = self.get(scanned_identifier)

                if not element:
                    continue

                # -- Get the tags for this element
                expected_tags = element.mandatory_tags()

                # -- Add them to our tracking variables
                all_tags.extend(expected_tags)
                mapped_tags[scanned_identifier] = expected_tags

            # -- Now we add all those tags into the db under a single
            # -- connection
            with ConnectionContext(self.identifier, commit=True) as ctxt:
                for tag in set(all_tags):
                    self._execute(
                        ctxt,
                        'tag_insert',
                        replacements={
                            '$(TAG)': tag,
                        },
                    )

            # -- With the tags all added, we now create the links between
            # -- the tags and the identifiers.
            with ConnectionContext(self.identifier, commit=True) as ctxt:
                for scanned_identifier in mapped_tags:
                    for tag in mapped_tags[scanned_identifier]:
                        try:
                            self._execute(
                                ctxt,
                                'tag_connect',
                                replacements={
                                    '$(IDENTIFIER)': scanned_identifier,
                                    '$(TAG)': tag,
                                },
                            )
                        except sqlite3.IntegrityError:
                            pass

            # -- If the project allows for data cleanup, lets perform
            # -- that now
            for identifier in self.find(None):
                for scan_plugin in self._scan_factory.plugins():
                    if scan_plugin.check(identifier) == scan_plugin.NOT_VALID:
                        self.remove(identifier)
                        break

        # -- Emit the fact that the scan is now complete
        self.scan_complete.emit()

        # -- Log the completion
        constants.log.info('Scan Complete : %s' % locations)

        return scanned_identifiers

    # --------------------------------------------------------------------------
    def explore(self, location):
        """
        This will return the above and below for the given location

        :param location:
        :return:
        """
        # -- Cycle over all our scan plugins
        for plugin in self._scan_factory.plugins():
            if plugin.can_represent(location):
                return plugin.above(location), plugin.below(location)

        return list(), list()

    # --------------------------------------------------------------------------
    def get(self, identifier):
        """
        This will create a composite binding of a DataElement plugin, bringing
        together all the plugins which can viably represent the data.

        :param identifier: Data Identifier to be passed to the DataElement
            plugins.
        :type identifier: str

        :return: DataElement composite 
        """
        template = None

        # -- Cycle all the data plugins and composite them together
        # -- if they can represent the given data
        for data_plugin in self._data_plugins:
            if data_plugin.can_represent(identifier):

                template = template or DataElement(
                    identifier,
                    self,
                )

                # -- Bind this plugin to the template
                template.bind(data_plugin(identifier, self))

        # -- If in debug mode, log the result
        constants.log.debug('Compounded %s to %s' % (identifier, template))

        return template

    # --------------------------------------------------------------------------
    def _execute(self, ctxt, command, replacements=None):
        """
        Private method which all sql queries should be routed through. This
        ensures a consistent result and suite of reporting.

        :param ctxt: ConnectionContext (all _execute calls should be done
            within a ConnectionContext to aid performance).

        :param command: The sql command name to run

        :param replacements: Any search and replace strings to process
            on the sql command
        :return:
        """
        # -- Get the statement
        statements = self._commands[command]

        # -- We must execute our sql in individual statements
        for single_statement in statements:

            # -- Create the replacements
            if replacements:
                for k, v in replacements.items():
                    single_statement = single_statement.replace(k, str(v))

            # -- Check if we need to print
            constants.log.debug('\n' + ('-' * 100))
            constants.log.debug(single_statement)

            try:
                ctxt.cursor.execute(single_statement)

            except sqlite3.Error:
                constants.log.debug(sys.exc_info())
                constants.log.debug('Unable to execute : %s' % single_statement)
                return list()

    # --------------------------------------------------------------------------
    def _command_dict(self):
        """
        This generates a dictionary whereby the key is the sql command name
        and the value is the sql data.

        :return: dict
        """
        command_dict = dict()

        for command in os.listdir(self._COMMAND_DIR):
            with open(os.path.join(self._COMMAND_DIR, command)) as f:
                command_dict[command[:-4]] = [
                    statement.strip()
                    for statement in f.read().split(';')
                    if statement.strip()
                ]
        return command_dict

    # --------------------------------------------------------------------------
    @classmethod
    def _conform(cls, item):
        """
        Takes in a path and ensures its correctly conformed for use
        within fracture.

        :param item: str

        :return: str
        """
        return item.replace('\\', '/')


# ------------------------------------------------------------------------------
class ConnectionContext(object):
    """
    This is a contextual class which handles the connection and cursor
    mechanisms. Its intended to be an optimisation to allow for multiple
    calls to occur without connection overheads.
    """
    # --------------------------------------------------------------------------
    def __init__(self, identifier, commit=False, get=False):
        self.identifier = identifier
        self.connection = None
        self.cursor = None
        self.results = []

        self.commit = commit
        self.get = get

    # --------------------------------------------------------------------------
    def __enter__(self):

        self.connection = sqlite3.connect(self.identifier)
        self.cursor = self.connection.cursor()
        return self

    # --------------------------------------------------------------------------
    def __exit__(self, exc_type, exc_val, exc_tb):

        # -- If we need to get, call a fetch
        if self.get:
            self.results = self.cursor.fetchall()

        if self.commit:
            self.connection.commit()

        self.connection.close()


# ------------------------------------------------------------------------------
def icon(icon_name='fracture'):
    """
    Returns the fracture icon

    :param icon_name: Name of the icon you want to request
    :return:
    """
    return os.path.join(
        os.path.dirname(__file__),
        '_res',
        '%s.png' % icon_name,
    )
